import json
from typing import Any, Dict, Optional, Sequence, Tuple, Type, Union

from bitarray import bitarray

from pyais.ais_types import AISType
from pyais.constants import TalkerID
from pyais.decode import decode
from pyais.exceptions import InvalidNMEAMessageException
from pyais.util import decode_into_bit_array, get_int, compute_checksum, deprecated


def validate_message(msg: bytes) -> None:
    """
    Validates a given message.
    It checks if the messages complies with the AIS standard.
    It is based on:
        1. https://en.wikipedia.org/wiki/Automatic_identification_system
        2. https://en.wikipedia.org/wiki/NMEA_0183

    If not errors are found, nothing is returned.
    Otherwise an InvalidNMEAMessageException is raised.
    """
    values = msg.split(b",")

    # A message has exactly 7 comma separated values
    if len(values) != 7:
        raise InvalidNMEAMessageException(
            "A NMEA message needs to have exactly 7 comma separated entries."
        )

    # The only allowed blank value may be the message ID
    if not values[0]:
        raise InvalidNMEAMessageException(
            "The NMEA message type is empty!"
        )

    if not values[1]:
        raise InvalidNMEAMessageException(
            "Number of sentences is empty!"
        )

    if not values[2]:
        raise InvalidNMEAMessageException(
            "Sentence number is empty!"
        )

    if not values[5]:
        raise InvalidNMEAMessageException(
            "The NMEA message body (payload) is empty."
        )

    if not values[6]:
        raise InvalidNMEAMessageException(
            "NMEA checksum (NMEA 0183 Standard CRC16) is empty."
        )

    try:
        sentence_num = int(values[1])
        if sentence_num > 9:
            raise InvalidNMEAMessageException(
                "Number of sentences exceeds limit of 9 total sentences."
            )
    except ValueError:
        raise InvalidNMEAMessageException(
            "Invalid sentence number. No Number."
        )

    if values[2]:
        try:
            sentence_num = int(values[2])
            if sentence_num > 9:
                raise InvalidNMEAMessageException(
                    " Sentence number exceeds limit of 9 total sentences."
                )
        except ValueError:
            raise InvalidNMEAMessageException(
                "Invalid Sentence number. No Number."
            )

    if values[3]:
        try:
            sentence_num = int(values[3])
            if sentence_num > 9:
                raise InvalidNMEAMessageException(
                    "Number of sequential message ID exceeds limit of 9 total sentences."
                )
        except ValueError:
            raise InvalidNMEAMessageException(
                "Invalid  sequential message ID. No Number."
            )

    # It should not have more than 82 chars of payload
    if len(values[5]) > 82:
        raise InvalidNMEAMessageException(
            f"{msg.decode('utf-8')} has more than 82 characters of payload."
        )

    # Only encapsulated messages are currently supported
    if values[0][0] != 0x21:
        # https://en.wikipedia.org/wiki/Automatic_identification_system
        raise InvalidNMEAMessageException(
            "'NMEAMessage' only supports !AIVDM/!AIVDO encapsulated messages. "
            f"These start with an '!', but got '{chr(values[0][0])}'"
        )


class NMEAMessage(object):
    __slots__ = (
        'ais_id',
        'raw',
        'talker',
        'type',
        'message_fragments',
        'fragment_number',
        'message_id',
        'channel',
        'payload',
        'fill_bits',
        'checksum',
        'bit_array'
    )

    def __init__(self, raw: bytes) -> None:
        if not isinstance(raw, bytes):
            raise ValueError(f"'NMEAMessage' only accepts bytes, but got '{type(raw)}'")

        validate_message(raw)

        # Initial values
        self.checksum: int = -1

        # Store raw data
        self.raw: bytes = raw

        # An AIS NMEA message consists of seven, comma separated parts
        values = raw.split(b",")

        # Unpack NMEA message parts
        (
            head,
            message_fragments,
            fragment_number,
            message_id,
            channel,
            payload,
            checksum
        ) = values

        # The talker is identified by the next 2 characters
        talker: str = head[1:3].decode('ascii')
        self.talker: TalkerID = TalkerID(talker)

        # The type of message is then identified by the next 3 characters
        self.type: str = head[3:].decode('ascii')

        # Total number of fragments
        self.message_fragments: int = int(message_fragments)
        # Current fragment index
        self.fragment_number: int = int(fragment_number)
        # Optional message index for multiline messages
        self.message_id: Optional[int] = int(message_id) if message_id else None
        # Channel (A or B)
        self.channel: str = channel.decode('ascii')
        # Decoded message payload as byte string
        self.payload: bytes = payload
        # Fill bits (0 to 5)
        self.fill_bits: int = int(chr(checksum[0]))
        # Message Checksum (hex value)
        self.checksum = int(checksum[2:], 16)

        # Finally decode bytes into bits
        self.bit_array: bitarray = decode_into_bit_array(self.payload)
        self.ais_id: int = get_int(self.bit_array, 0, 6)

    def __str__(self) -> str:
        return str(self.raw)

    def __getitem__(self, item: str) -> Union[int, str, bytes, bitarray]:
        if isinstance(item, str):
            try:
                return getattr(self, item)  # type: ignore
            except AttributeError:
                raise KeyError(item)
        else:
            raise TypeError(f"Index must be str, not {type(item).__name__}")

    def asdict(self) -> Dict[str, Any]:
        """
        Convert the class to dict.
        @return: A dictionary that holds all fields, defined in __slots__
        """
        return {
            'ais_id': self.ais_id,  # int
            'raw': self.raw.decode('ascii'),  # str
            'talker': self.talker.value,  # str
            'type': self.type,  # str
            'message_fragments': self.message_fragments,  # int
            'fragment_number': self.fragment_number,  # int
            'message_id': self.message_id,  # None or int
            'channel': self.channel,  # str
            'payload': self.payload.decode('ascii'),  # str
            'fill_bits': self.fill_bits,  # int
            'checksum': self.checksum,  # int
            'bit_array': self.bit_array.to01(),  # str
        }

    def __eq__(self, other: object) -> bool:
        return all([getattr(self, attr) == getattr(other, attr) for attr in self.__slots__])

    @classmethod
    def from_string(cls, nmea_str: str) -> "NMEAMessage":
        return cls(nmea_str.encode('utf-8'))

    @classmethod
    def from_bytes(cls, nmea_byte_str: bytes) -> "NMEAMessage":
        return cls(nmea_byte_str)

    @classmethod
    def assemble_from_iterable(cls, messages: Sequence["NMEAMessage"]) -> "NMEAMessage":
        """
        Assemble a multiline message from a sequence of NMEA messages.
        :param messages: Sequence of NMEA messages
        :return: Single message
        """
        raw = b''
        data = b''
        bit_array = bitarray()

        for msg in messages:
            raw += msg.raw
            data += msg.payload
            bit_array += msg.bit_array

        messages[0].raw = raw
        messages[0].payload = data
        messages[0].bit_array = bit_array
        return messages[0]

    @property
    def is_valid(self) -> bool:
        return self.checksum == compute_checksum(self.raw)

    @property
    def is_single(self) -> bool:
        return not self.message_id and self.fragment_number == self.message_fragments == 1

    @property
    def is_multi(self) -> bool:
        return not self.is_single

    @property
    def fragment_count(self) -> int:
        return self.message_fragments

    @deprecated
    def count(self) -> int:
        """
        Only there fore legacy compatibility.
        Was renamed to `message_fragments`
        @return: message_fragments as int
        """
        return self.message_fragments

    @deprecated
    def index(self) -> int:
        """
        Only there fore legacy compatibility.
        Was renamed to `fragment_number`
        @return:  fragment_number as int
        """
        return self.fragment_number

    @deprecated
    def seq_id(self) -> Optional[int]:
        """
        Only there fore legacy compatibility.
        Was renamed to `message_id`
        @return: message_id as int
        """
        return self.message_id

    @deprecated
    def data(self) -> bytes:
        """
        Only there fore legacy compatibility.
        Was renamed to `payload`
        @return: payload as sequence of bytes
        """
        return self.payload

    def decode(self, silent: bool = True) -> Optional["AISMessage"]:
        """
        Decode the message content.

        @param silent: Boolean. If set to true errors are ignored and None is returned instead
        """
        msg = AISMessage(self)
        try:
            msg.decode()
        except Exception as e:
            if not silent:
                raise e

        return msg


class AISMessage(object):
    """
    Initializes a generic AIS message.
    """

    def __init__(self, nmea_message: NMEAMessage) -> None:
        """Creates an initial empty AIS message"""
        self.nmea: NMEAMessage = nmea_message
        self.msg_type: AISType = AISType.NOT_IMPLEMENTED
        self.content: Dict[str, Any] = {}

    def __getitem__(self, item: str) -> Any:
        return self.content[item]

    def __str__(self) -> str:
        return str(self.content)

    @property
    def fields(self) -> Tuple[Tuple[str, Type[Any]], ...]:
        return tuple([(str(key), type(value)) for (key, value) in self.content.items()])

    def decode(self) -> None:
        """Decodes the given message and extracts it's type and content.
        This function potentially fails, if the message is malformed."""
        self.msg_type = AISType(self.nmea.ais_id)
        self.content = decode(self.nmea)

    def asdict(self) -> Dict[str, Any]:
        return {
            'nmea': self.nmea.asdict(),
            'decoded': self.content
        }

    def to_json(self) -> str:
        return json.dumps(
            self.asdict(),
            indent=4
        )



import pathlib
import unittest
from unittest.case import skip
from pyais.stream import FileReaderStream, should_parse
from pyais.messages import NMEAMessage


class TestFileReaderStream(unittest.TestCase):
    FILENAME = "tests/ais_test_messages"

    def test_reader(self):
        with FileReaderStream(self.FILENAME) as stream:
            messages = [msg for msg in stream]

        assert len(messages) == 7
        for msg in messages:
            assert type(msg) == NMEAMessage
            assert msg.is_valid
            assert msg.decode().content is not None

    def test_reader_with_open(self):
        with FileReaderStream(self.FILENAME) as stream:
            msg = next(stream)
            assert type(msg) == NMEAMessage
            assert msg.is_valid
            assert msg.decode().content is not None

    def test_invalid_filename(self):
        with self.assertRaises(FileNotFoundError):
            FileReaderStream("doesnotexist")

    @skip("This takes too long for now")
    def test_large_file(self):
        # The ais sample data is downloaded from https://www.aishub.net/ais-dispatcher
        par_dir = pathlib.Path(__file__).parent.absolute()
        large_file = par_dir.joinpath("nmea-sample")
        for msg in FileReaderStream(large_file):
            msg.decode()

    def test_marine_traffic_sample(self):
        """Test some messages from https://help.marinetraffic.com/hc/en-us
        /articles/215626187-I-am-an-AIS-data-contributor-Can-you-share-more-data-with-me-"""

        par_dir = pathlib.Path(__file__).parent.absolute()
        nmea_file = par_dir.joinpath("nmea_data_sample.txt")

        with FileReaderStream(nmea_file) as stream:
            for msg in stream:
                assert msg.decode()

    def test_should_parse(self):
        self.assertFalse(should_parse(b""))
        self.assertFalse(should_parse(b"#"))
        self.assertFalse(should_parse(b"$"))
        self.assertFalse(should_parse(b"!"))
        self.assertFalse(should_parse(b"Foo"))
        self.assertFalse(should_parse(b"!Baz"))
        self.assertFalse(should_parse(b"$fooBar"))
        self.assertFalse(should_parse(b"Foo,Bar,Baz,Foo,Bar,Baz"))
        self.assertFalse(should_parse(b"Foo,Bar,Baz,Foo,Bar,Baz,Foo"))
        self.assertFalse(should_parse(b"$Foo,Bar,Baz,Foo,Bar,Baz,Foo,"))
        self.assertFalse(should_parse(b"!Foo,Bar,Baz,Foo,Bar,Baz,Foo,"))

        self.assertTrue(should_parse(b"!,,,,,,"))
        self.assertTrue(should_parse(b"!AIVDM,1,1,,B,,0*25"))
        self.assertTrue(should_parse(b"$AIVDM,1,1,,B,,0*25"))
        self.assertTrue(should_parse(b"$Foo,Bar,Baz,Foo,Bar,Baz,Foo"))
        self.assertTrue(should_parse(b"!AIVDM,1,1,,B,33K:7p001k1e5@TE?8rTH3BH011@,0*59"))
        self.assertTrue(should_parse(b"$AIVDM,1,1,,B,33K:7p001k1e5@TE?8rTH3BH011@,0*59"))
        self.assertTrue(should_parse(b"!AIVDM,2,1,9,B,53nFBv01SJ<thHp6220H4heHTf2222222222221?50:454o<`9QSlUDp,0*09"))
        self.assertTrue(should_parse(b"$AIVDM,2,1,9,B,53nFBv01SJ<thHp6220H4heHTf2222222222221?50:454o<`9QSlUDp,0*09"))

    def test_mixed_content(self):
        """Test that the file reader handles mixed content. That means, that is is able to handle
        text files, that contain both AIS messages and non AIS messages."""
        par_dir = pathlib.Path(__file__).parent.absolute()
        mixed_content_file = par_dir.joinpath("messages.ais")
        self.assertEqual(len(list(iter(FileReaderStream(mixed_content_file)))), 6)



from enum import IntEnum, Enum

# Keywords
UNDEFINED = 'Undefined'
RESERVED = 'Reserved'
NULL = 'N/A'
ANSI_RED = '\x1b[31m'
ANSI_RESET = '\x1b[0m'


class TalkerID(str, Enum):
    """ Enum of all  NMEA talker IDs.
    See: https://gpsd.gitlab.io/gpsd/AIVDM.html#_talker_ids"""
    Base_Station = "AB"
    Dependent_Base_Station = "AD"
    Mobile_Station = "AI"
    Navigation_Station = "AN"
    Receiving_Station = "AR"
    Limited_Base_Station = "AS"
    Transmitting_Station = "AT"
    Repeater_Station = "AX"
    Base_Station_Deprecated = "BS"
    Physical_Shore_Station = "SA"
    UNDEFINED = "UNDEFINED"

    @classmethod
    def _missing_(cls, value: object) -> str:
        return TalkerID.UNDEFINED


class NavigationStatus(IntEnum):
    UnderWayUsingEngine = 0
    AtAnchor = 1
    NotUnderCommand = 2
    RestrictedManoeuverability = 3
    ConstrainedByHerDraught = 4
    Moored = 5
    Aground = 6
    EngagedInFishing = 7
    UnderWaySailing = 8
    AISSARTActive = 14
    Undefined = 15

    @classmethod
    def _missing_(cls, value: object) -> int:
        return NavigationStatus.Undefined


class ManeuverIndicator(IntEnum):
    NotAvailable = 0
    NoSpecialManeuver = 1
    SpecialManeuver = 2
    UNDEFINED = 3

    @classmethod
    def _missing_(cls, value: object) -> int:
        return ManeuverIndicator.UNDEFINED


class EpfdType(IntEnum):
    Undefined = 0
    GPS = 1
    GLONASS = 2
    GPS_GLONASS = 3
    Loran_C = 4
    Chayka = 5
    IntegratedNavigationSystem = 6
    Surveyed = 7
    Galileo = 8

    @classmethod
    def _missing_(cls, value: object) -> int:
        return EpfdType.Undefined


class ShipType(IntEnum):
    NotAvailable = 0
    # 20's
    WIG = 20
    WIG_HazardousCategory_A = 21
    WIG_HazardousCategory_B = 22
    WIG_HazardousCategory_C = 23
    WIG_HazardousCategory_D = 24
    WIG_Reserved = 25
    # 30's
    Fishing = 30
    Towing = 31
    Towing_LengthOver200 = 32
    DredgingOrUnderwaterOps = 33
    DivingOps = 34
    MilitaryOps = 35
    Sailing = 36
    PleasureCraft = 37
    # 40's
    HSC = 40
    HSC_HazardousCategory_A = 41
    HSC_HazardousCategory_B = 42
    HSC_HazardousCategory_C = 43
    HSC_HazardousCategory_D = 44
    HSC_Reserved = 45
    HSC_NoAdditionalInformation = 49
    # 50's
    PilotVessel = 50
    SearchAndRescueVessel = 51
    Tug = 52
    PortTender = 53
    AntiPollutionEquipment = 54
    LawEnforcement = 55
    SPARE = 56
    MedicalTransport = 58
    NonCombatShip = 59
    # 60's
    Passenger = 60
    Passenger_HazardousCategory_A = 61
    Passenger_HazardousCategory_B = 62
    Passenger_HazardousCategory_C = 63
    Passenger_HazardousCategory_D = 64
    Passenger_Reserved = 65
    Passenger_NoAdditionalInformation = 69
    # 70's
    Cargo = 70
    Cargo_HazardousCategory_A = 71
    Cargo_HazardousCategory_B = 72
    Cargo_HazardousCategory_C = 73
    Cargo_HazardousCategory_D = 74
    Cargo_Reserved = 75
    Cargo_NoAdditionalInformation = 79
    # 80's
    Tanker = 80
    Tanker_HazardousCategory_A = 81
    Tanker_HazardousCategory_B = 82
    Tanker_HazardousCategory_C = 83
    Tanker_HazardousCategory_D = 84
    Tanker_Reserved = 85
    Tanker_NoAdditionalInformation = 89
    # 90's
    OtherType = 90
    OtherType_HazardousCategory_A = 91
    OtherType_HazardousCategory_B = 92
    OtherType_HazardousCategory_C = 93
    OtherType_HazardousCategory_D = 94
    OtherType_Reserved = 95
    OtherType_NoAdditionalInformation = 99

    @classmethod
    def _missing_(cls, value: object) -> int:
        if isinstance(value, int):
            if 24 < value < 30:
                return ShipType.WIG_Reserved

            if 44 < value < 49:
                return ShipType.HSC_Reserved

            if 55 < value < 58:
                return ShipType.SPARE

            if 64 < value < 69:
                return ShipType.Passenger_Reserved

            if 74 < value < 79:
                return ShipType.Cargo_Reserved

            if 84 < value < 89:
                return ShipType.Tanker_Reserved

            if 94 < value < 99:
                return ShipType.OtherType_Reserved

        return ShipType.NotAvailable


class DacFid(IntEnum):
    DangerousCargoIndication = 13
    TidalWindow = 15
    NumPersonsOnBoard = 17
    ClearanceTimeToEnterPort = 19
    BerthingData = 21
    AreaNotice = 24
    RouteInfoAddressed = 29
    TextDescriptionAddressed = 31
    ETA = 221
    RTA = 222
    AtoN_MonitoringData_UK = 245
    AtoN_MonitoringData_ROI = 260


class NavAid(IntEnum):
    DEFAULT = 0
    REFERENCE_POINT = 1
    RACON = 2
    FIXED = 3
    FITTED = 4
    SPARE = 5
    LIGHT_NO_SECTORS = 6
    LIGHT_SECTORS = 7
    LEADING_LIGHT_FRONT = 8
    LEADING_LIGHT_REAR = 9
    BEACON_CARDINAL_N = 10
    BEACON_CARDINAL_E = 11
    BEACON_CARDINAL_S = 12
    BEACON_CARDINAL_W = 13
    BEACON_STARBOARD = 14
    BEACON_CHANNEL_PORT_HAND = 15
    BEACON_CHANNEL_STARBOARD_HAND = 16
    BEACON_ISOLATED_DANGER = 17
    BEACON_SAFE_WATER = 18
    BEACON_SPECIAL_MARK = 19
    CARDINAL_MARK_N = 20
    CARDINAL_MARK_E = 21
    CARDINAL_MARK_S = 22
    CARDINAL_MARK_W = 23
    PORT_HAND_MARK = 24
    STARBOARD_HAND_MARK = 25
    PREFERRED_HAND_PORT_HAND = 26
    PREFERRED_HAND_STARBOARD_HAND = 27
    ISOLATED_DANGER = 28
    SAFE_WATER = 29
    SPECIAL_MARK = 30
    LIGHT_VESSEL = 31

    @classmethod
    def _missing_(cls, value: object) -> int:
        return NavAid.DEFAULT


class TransmitMode(IntEnum):
    TXA_TXB_RXA_RXB = 0  # default
    TXA_RXA_RXB = 1
    TXB_RXA_RXB = 2
    RESERVED = 3

    @classmethod
    def _missing_(cls, value: object) -> int:
        return TransmitMode.TXA_TXB_RXA_RXB


class StationType(IntEnum):
    ALL = 0
    RESERVED = 1
    CLASS_B_ALL = 2
    SAR_AIRBORNE = 3
    AID_NAV = 4
    CLASS_B_SHIPBORNE = 5
    REGIONAL = 6

    @classmethod
    def _missing_(cls, value: object) -> int:
        if isinstance(value, int):
            if 6 <= value <= 9:
                return StationType.REGIONAL
            if 10 <= value <= 15:
                return StationType.RESERVED
        return StationType.ALL


class StationIntervals(IntEnum):
    AUTONOMOUS_MODE = 0
    MINUTES_10 = 1
    MINUTES_6 = 2
    MINUTES_3 = 3
    MINUTES_1 = 4
    SECONDS_30 = 5
    SECONDS_15 = 6
    SECONDS_10 = 7
    SECONDS_5 = 8
    NEXT_SHORT_REPORTER_INTERVAL = 9
    NEXT_LONGER_REPORTING_INTERVAL = 10
    RESERVED = 11

    @classmethod
    def _missing_(cls, value: object) -> int:
        return StationIntervals.RESERVED



class InvalidNMEAMessageException(Exception):
    """Invalid NMEA Message"""
    pass


class UnknownMessageException(Exception):
    """Message not supported yet"""
    pass


class MissingMultipartMessageException(Exception):
    """Multipart message with missing parts provided"""


class TooManyMessagesException(Exception):
    """Too many messages"""



import warnings
from collections import OrderedDict
from functools import partial, reduce
from operator import xor
from typing import Any, Generator, Hashable, TYPE_CHECKING, Callable

from bitarray import bitarray

if TYPE_CHECKING:
    BaseDict = OrderedDict[Hashable, Any]
else:
    BaseDict = OrderedDict

from_bytes = partial(int.from_bytes, byteorder="big")
from_bytes_signed = partial(int.from_bytes, byteorder="big", signed=True)


def deprecated(f: Callable[[Any], Any]) -> Callable[[Any], Any]:
    @property  # type: ignore
    def wrapper(self: Any) -> Any:
        warnings.simplefilter('always', DeprecationWarning)  # turn off filter
        warnings.warn(f"{f.__name__} is deprecated and will be removed soon.",
                      category=DeprecationWarning)
        warnings.simplefilter('default', DeprecationWarning)  # reset filter

        return f(self)

    return wrapper


def decode_into_bit_array(data: bytes) -> bitarray:
    """
    Decodes a raw AIS message into a bitarray.
    :param data: Raw AIS message in bytes, as it is received from a TCP socket.
    :return:
    """
    bit_arr = bitarray()

    for _, c in enumerate(data):
        if c < 0x30 or c > 0x77 or 0x57 < c < 0x6:
            raise ValueError(f"Invalid character: {chr(c)}")

        # Convert 8 bit binary to 6 bit binary
        c -= 0x30 if (c < 0x60) else 0x38
        c &= 0x3F
        bit_arr += bitarray(f'{c:06b}')

    return bit_arr


def chunks(sequence: bitarray, n: int) -> Generator[bitarray, None, None]:
    """Yield successive n-sized chunks from l."""
    return (sequence[i:i + n] for i in range(0, len(sequence), n))


def encode_bin_as_ascii6(bit_arr: bitarray) -> str:
    """
    Encode binary data as 6 bit ASCII.
    :param bit_arr: array of bits
    :return: ASCII String
    """
    string: str = ""
    for c in chunks(bit_arr, 6):
        n: int = from_bytes(c.tobytes()) >> 2

        # Last entry may not have 6 bits
        if len(c) != 6:
            n >> (6 - len(c))

        if n < 0x20:
            n += 0x40

        # Break if there is an @
        if n == 64:
            break

        string += chr(n)

    return string.strip()


def get_int(data: bitarray, ix_low: int, ix_high: int, signed: bool = False) -> int:
    """
    Cast a subarray of a bitarray into an integer.
    The bitarray module adds tailing zeros when calling tobytes(), if the bitarray is not a multiple of 8.
    So those need to be shifted away.
    :param data: some bitarray
    :param ix_low: the lower index of the sub-array
    :param ix_high: the upper index of the sub-array
    :param signed: True if the value should be interpreted as a signed integer
    :return: a normal integer (int)
    """
    shift: int = (8 - ((ix_high - ix_low) % 8)) % 8
    data = data[ix_low:ix_high]
    i: int = from_bytes_signed(data) if signed else from_bytes(data)
    return i >> shift


def get_mmsi(data: bitarray, ix_low: int, ix_high: int) -> str:
    """
    A Maritime Mobile Service Identity (MMSI) is a series of nine digits.
    Every digit is required and therefore we can NOT use a int.
    See: issue #6
    """
    mmsi_int: int = get_int(data, ix_low, ix_high)
    return str(mmsi_int).zfill(9)


def compute_checksum(msg: bytes) -> int:
    """
    Compute the checksum of a given message
    :param msg: message
    :return: hex
    """
    msg = msg[1:].split(b'*', 1)[0]
    return reduce(xor, msg)


class FixedSizeDict(BaseDict):
    """
    Fixed sized dictionary that only contains up to N keys.
    """

    def __init__(self, maxlen: int) -> None:
        super().__init__()
        self.maxlen: int = maxlen

    def __setitem__(self, k: Hashable, v: Any) -> None:
        super().__setitem__(k, v)
        # if the maximum number is reach delete the oldest n keys
        if len(self) >= self.maxlen:
            self._pop_oldest()

    def _pop_oldest(self) -> Any:
        # instead of calling this method often, we delete a whole bunch of keys in one run
        for _ in range(self._items_to_pop):
            self.popitem(last=False)

    @property
    def _items_to_pop(self) -> int:
        # delete 1/5th of keys
        return self.maxlen // 5



from abc import ABC, abstractmethod
from socket import AF_INET, SOCK_DGRAM, SOCK_STREAM, socket
from typing import (
    Any, BinaryIO, Generator, Generic, Iterable, List, Optional, TypeVar, cast
)

from pyais.exceptions import InvalidNMEAMessageException
from pyais.messages import NMEAMessage
from pyais.util import FixedSizeDict

F = TypeVar("F", BinaryIO, socket, None)
DOLLAR_SIGN = ord("$")
EXCLAMATION_POINT = ord("!")


def should_parse(byte_str: bytes) -> bool:
    """Return True if a given byte string seems to be NMEA message.
    This method does **NOT** validate the message, but uses a heuristic
    approach to check (or guess) if byte string is a valid nmea_message.
    """
    # The byte sequence is not empty and starts with a $ or a ! and has 6 ','
    return len(byte_str) > 0 and byte_str[0] in (DOLLAR_SIGN, EXCLAMATION_POINT) and byte_str.count(b",") == 6


class AssembleMessages(ABC):
    """
    Base class that assembles multiline messages.
    Offers a iterator like interface.

    This class comes without a __init__ method, because it should never be instantiated!
    """

    def __enter__(self) -> "AssembleMessages":
        # Enables use of with statement
        return self

    def __exit__(self, exc_type: object, exc_val: object, exc_tb: object) -> None:
        return None

    def __iter__(self) -> Generator[NMEAMessage, None, None]:
        return self._assemble_messages()

    def __next__(self) -> NMEAMessage:
        """Returns the next decoded NMEA message."""
        return next(iter(self))

    def _assemble_messages(self) -> Generator[NMEAMessage, None, None]:
        queue: List[NMEAMessage] = []

        for line in self._iter_messages():

            # Be gentle and just skip invalid messages
            try:
                msg: NMEAMessage = NMEAMessage(line)
            except InvalidNMEAMessageException:
                continue

            if msg.is_single:
                yield msg

            # Assemble multiline messages
            elif msg.is_multi:
                queue.append(msg)

                if msg.fragment_number == msg.message_fragments:
                    yield msg.assemble_from_iterable(queue)
                    queue.clear()
            else:
                raise ValueError("Messages are out of order!")

    @abstractmethod
    def _iter_messages(self) -> Generator[bytes, None, None]:
        raise NotImplementedError("Implement me!")


class IterMessages(AssembleMessages):

    def __init__(self, messages: Iterable[bytes]):
        # If the user passes a single byte string make it into a list
        if isinstance(messages, bytes):
            messages = [messages, ]
        self.messages: Iterable[bytes] = messages

    @classmethod
    def from_strings(cls, messages: Iterable[str], ignore_encoding_errors: bool = False,
                     encoding: str = "utf-8") -> "IterMessages":
        # If the users passes a single message as string, make it a list
        if isinstance(messages, str):
            messages = [messages, ]

        encoded: List[bytes] = []
        for message in messages:
            try:
                encoded.append(message.encode(encoding))
            except UnicodeEncodeError as e:
                if ignore_encoding_errors:
                    # Just skip and carry on
                    continue
                raise e

        return IterMessages(encoded)

    def _iter_messages(self) -> Generator[bytes, None, None]:
        # Transform self.messages into a generator
        yield from (message for message in self.messages)


class Stream(AssembleMessages, Generic[F], ABC):

    def __init__(self, fobj: F) -> None:
        """
        Create a new Stream-like object.
        @param fobj: A file-like or socket object.
        """
        self._fobj: F = fobj

    def __exit__(self, exc_type: object, exc_val: object, exc_tb: object) -> None:
        if self._fobj is not None:
            self._fobj.close()

    def _iter_messages(self) -> Generator[bytes, None, None]:
        # Do not parse lines, that are obviously not NMEA messages
        yield from (line for line in self.read() if should_parse(line))

    @abstractmethod
    def read(self) -> Generator[bytes, None, None]:
        raise NotImplementedError()


class BinaryIOStream(Stream[BinaryIO]):
    """Read messages from a file-like object"""

    def __init__(self, file: BinaryIO) -> None:
        super().__init__(file)

    def read(self) -> Generator[bytes, None, None]:
        yield from self._fobj.readlines()


class FileReaderStream(BinaryIOStream):
    """
    Read NMEA messages from file
    """

    def __init__(self, filename: str, mode: str = "rb") -> None:
        self.filename: str = filename
        self.mode: str = mode
        # Try to open file
        try:
            file = open(self.filename, mode=self.mode)
            file = cast(BinaryIO, file)
        except Exception as e:
            raise FileNotFoundError(f"Could not open file {self.filename}") from e
        super().__init__(file)


class ByteStream(Stream[None]):
    """
    Takes a iterable that contains ais messages as bytes and assembles them.
    """

    def __init__(self, iterable: Iterable[bytes]) -> None:
        self.iterable: Iterable[bytes] = iterable
        super().__init__(None)

    def __exit__(self, exc_type: object, exc_val: object, exc_tb: object) -> None:
        return

    def read(self) -> Generator[bytes, None, None]:
        yield from self.iterable


class OutOfOrderByteStream(Stream[F], ABC):
    """
    Handles multipart NMEA that are delivered out of order.

    This class is not attached to a datasource by default.
    You need to subclass it and override _get_messages().
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        # create a fixed sized message queue
        self._queue = FixedSizeDict(10000)
        super().__init__(*args, **kwargs)

    def __exit__(self, exc_type: object, exc_val: object, exc_tb: object) -> None:
        del self._queue
        super().__exit__(exc_type, exc_val, exc_tb)

    def _split_nmea_header(self, msg: bytes) -> None:
        """
        Read the important parts of a NMEA header
        """
        parts: List[bytes] = msg.split(b',')
        self.seq_id: int = int(parts[3]) if parts[3] else 0
        self.fragment_offset: int = int(parts[2]) - 1
        self.fragment_count: int = int(parts[1])

    def _yield_complete(self) -> Optional[List[bytes]]:
        """
        Check if the message is complete and return it
        """
        # get all messages for the current sequence number
        queue: List[bytes] = self._queue[self.seq_id][0:self.fragment_count]
        if all(queue[0: self.fragment_count]):
            # if all required messages are received yield them and free their space
            del self._queue[self.seq_id]
            return queue[0: self.fragment_count]
        return None

    def _add_to_queue(self, msg: bytes) -> None:
        """
        Append a new nmea message to queue
        """
        # MAX frag offset for any AIS NMEA is 9
        msg_queue: List[Optional[bytes]] = ([None, ] * 9)
        try:
            # place the message at its correct position
            msg_queue[self.fragment_offset] = msg
        except IndexError:
            # message is invalid clear it
            del self._queue[self.seq_id]
        self._queue[self.seq_id] = msg_queue

    def _update_queue(self, msg: bytes) -> Optional[Iterable[bytes]]:
        """
        Update an existing message queue that is not complete yet.
        Return a list of fully assembled messages if all required messages for a given sequence number are received.
        """
        msg_queue: List[bytes] = self._queue[self.seq_id]
        msg_queue[self.fragment_offset] = msg
        complete: Optional[List[bytes]] = self._yield_complete()
        if complete:
            yield from complete
        return None

    def _iter_messages(self) -> Generator[bytes, None, None]:
        for msg in self.read():
            # decode nmea header
            self._split_nmea_header(msg)
            if not self.seq_id:
                yield msg
            try:
                complete = self._update_queue(msg)
                if complete is not None:
                    yield from complete

            except KeyError:
                # place item a correct pos and then store the list
                self._add_to_queue(msg)


class SocketStream(Stream[socket]):
    BUF_SIZE = 4096

    def read(self) -> Generator[bytes, None, None]:
        partial: bytes = b''
        while True:
            body = self._fobj.recv(self.BUF_SIZE)
            # Server closed connection
            if not body:
                return None

            lines = body.split(b'\r\n')

            line = partial + lines[0]
            if line:
                yield line

            yield from (line for line in lines[1:-1] if line)

            partial = lines[-1]


class UDPStream(OutOfOrderByteStream[socket], SocketStream):

    def __init__(self, host: str, port: int) -> None:
        sock: socket = socket(AF_INET, SOCK_DGRAM)
        sock.bind((host, port))
        super().__init__(sock)


class TCPStream(SocketStream):
    """
     NMEA0183 stream via socket. Refer to
     https://en.wikipedia.org/wiki/NMEA_0183
     """

    def __init__(self, host: str, port: int = 80) -> None:
        sock: socket = socket(AF_INET, SOCK_STREAM)
        try:
            sock.connect((host, port))
        except ConnectionRefusedError as e:
            sock.close()
            raise ConnectionRefusedError(f"Failed to connect to {host}:{port}") from e
        super().__init__(sock)
